// conditional compile switches
//
#define CARD_EXIT_FILLS_PARTIAL_BLOCK   TRUE
#define CARD_EXIT_RELEASES_BLOCKS       TRUE
#define PENDING_BLOCK_PARANOIA          TRUE
#define AMORTIZE_PCI_BUS_CYCLES         TRUE
#define ATTEMPT_UNAMORTIZE              FALSE       // maybe unsafe, don't do this yet...



#include <stdio.h>                      // printf()
#include <stdlib.h>                     // atoi() strtoul()
#include "../pci/pci.h"                 // pci_get_irq_level()
#include "../apptel/driver.h"           // class AtmCardRunning
#include "../dpmi32/intr.h"             // getvect() setvect()
#include "../dpmi32/flat.h"             // flat_alloc() flat_to_physical() align()
#include "../apptel/sram.h"             // sram_init_stats() sram_print_stats() sram_read_or_write()
#include "../integrat/constant.h"       // CARD_COPIES_xxx CELLS_PER_PACKET



// must declare class static variables
// so they get allocated and initialized
//
InterruptController AtmCardRunning::irq;
int AtmCardRunning::card_count = 0;



AtmCardRunning::AtmCardRunning( Bit8 card_index0 )
:
    AtmCard( card_index0 )
{
    unsigned count;
    Bit32 sram_data = 0;

    // fix up the index to be zero-based,
    // now that we are attached to the correct card
    //
    card_index -= ATM_NIC_OFFSET;

    // first make sure that our base class initialized OK
    //
    if( init_result )
        return;

    // set state for all VPI/VCI's to "no cells seen yet"
    //
    // assumes that FPGA's are already downloaded
    //
    // we must do this before calling interrupt_init(),
    // so the destructor can always decide whether to call
    // interrupt_exit() by looking at init_result
    //
    printf( "card %d clearing VPI/VCI state memory\n", card_index );
    for( count=0; count<MAX_SRAM_ADDRESS; count++ )
    {
        init_result = sram_read_or_write( &both_fpgas->Receive, count, TRUE, &sram_data );
        if( init_result )
            return;
    }

    // now take over the interrupt handling
    //
    interrupt_init();
    if( init_result )
        return;

    // initialize our handling of SRAM
    //
    if( card_count == 0 )
        sram_init_stats();

    // update reference count
    //
    card_count++;

    // remember that card_init() hasn't been called yet
    //
    card_was_initialized = FALSE;
}


AtmCardRunning::~AtmCardRunning()
{
    // first make sure that we initialized OK
    //
    if( init_result )
        return;

    if( card_was_initialized )
        card_exit();

    card_count--;
    if( card_count == 0 )
        sram_print_stats();
    interrupt_exit();
}


// re-route the card interrupt to our code
//
void AtmCardRunning::interrupt_init( void )
{
    interrupt_result = 0;

    // get the interrupt assignments
    //
    irq_level = pci_get_irq_level( card_handle );
    interrupt_number = irq_to_intr( irq_level );
    old_handler = getvect( interrupt_number );
    fprintf( stdout, "card is using IRQ level %d and interrupt 0x%.2X\n", irq_level, interrupt_number );
    fprintf( stdout, "which is currently being handled by the function at 0x%.4X:%.8X\n", old_handler.selector, old_handler.offset );

    disable();

    thunk = make_thunk( (ifp)interrupt_handler0, &old_handler, this );
    if( thunk == NULL )
    {
        fprintf( stderr, "fatal internal error!  cannot allocate interrupt handler thunk\n" );
        init_result = ERROR_POINT2_THUNK_ALLOC;
        return;
    }

    Bit48 pointer;
    pointer.offset = (Bit32) thunk;
    pointer.selector = _CS;
    setvect( interrupt_number, pointer );

    irq.unmask( irq_level );
    enable();
}



// no need to mask the interrupt for this card,
// since the InterruptController destructor will
// restore previous masking for all IRQ levels
//
void AtmCardRunning::interrupt_exit( void )
{
    disable();
    setvect( interrupt_number, old_handler );
    free( thunk );
    enable();
}



int AtmCardRunning::interrupt_handler( void )
{
    // track how many interrupts happened
    //
    interrupt_count++;

    // find out which interrupts occurred
    //
    amcc_intr = amcc->InterruptControlStatus;

    // find out which bus mastering is currently allowed
    //
    amcc_bus = amcc->BusMasterControlStatus;

    // specific handling for each sub-interrupt
    //
    if( amcc_intr.WriteTransferCompleted )
    {
        // do whatever is necessary to keep the card in buffers
        //
        int result = write_completed_handler();
        if( result && !interrupt_result )
            interrupt_result = result;
    }

    // tell the interface card that we handled
    // all possible pending interrupts, so it doesn't
    // try to interrupt us again for the same situation
    //
    // note that this covers read transfer completion,
    // incoming mailbox becoming full, outgoing mailbox
    // becoming empty, master abortion, and target abortion,
    // even though we have no specific handlers for any
    // of them
    //
#if 0
    * (Bit32 *) &amcc_intr = 0;
    amcc_intr.OutgoingMailBoxInterrupted = TRUE;
    amcc_intr.IncomingMailBoxInterrupted = TRUE;
    amcc_intr.WriteTransferCompleted = TRUE;
    amcc_intr.ReadTransferCompleted = TRUE;
    amcc_intr.MasterAborted = TRUE;
    amcc_intr.TargetAborted = TRUE;
#endif
    amcc->InterruptControlStatus = amcc_intr;

    // tell the PC motherboard interrupt controllers we are done
    //
    irq.acknowledge( irq_level );

    // tell the thunking code not to chain to the old handler
    //
    return 0;
}



// manages one BufferBlock which has 17K buffers
// in the steady-state my_block is being filled in
//
// however, the edge condition, where the card has completed
// filling my_block and started filling its external 4K word
// FIFO, is when host gets interrupted and this code gets called
//
//
//
// this code does not deal with buffer exhaustion, so if the
// host ever runs out of BufferBlock's, the card will too!
//
int AtmCardRunning::write_completed_handler( void )
{
    int result;
    BufferBlock *next_block;

    // disable bus master writes to host memory
    //
    amcc_bus.EnableWriteTransfer = FALSE;
    amcc->BusMasterControlStatus = amcc_bus;

    // if we can allocate another block...
    //
    result = all_buffers->allocate( &next_block );
    if( !result )
    {
        // tell the card about it
        //
        amcc->MasterWriteAddress = flat_to_physical( next_block->buffers );
        amcc->MasterWriteTransferCount = COMPL_RING_ENTRIES * BUFFER_SIZE;

        // re-enable bus master writes to host memory
        //
        amcc_bus.EnableWriteTransfer = TRUE;
        amcc->BusMasterControlStatus = amcc_bus;

        // remember which card will fill in the block
        //
        next_block->cardnum = card_handle;
        next_block->cardindex = card_index;
    }

    // if the card had a block already...
    //
    if( my_block )
    {
        // assume all of the buffers were filled
        //
        my_block->used = COMPL_RING_ENTRIES;

        // place my_block on global filled list
        //
        result = all_buffers->fill( my_block );
        if( result != 0 )
            return result;
    }

    // shift the block pointers
    //
    my_block = next_block;

    return result;
}



// called when BufferQueue stops being empty,
// in case this card was in need of a buffer
//
void AtmCardRunning::foreground_replenish( void )
{
    int result;

    // if this card will interrupt us again...
    //
    if( my_block )

        // no need for any extra effort on our part
        //
        return;

    // try to get a BufferBlock and give it to the card
    //
    // as soon as we give it the block, data buffered in the
    // external FIFO's will start coming in to host memory
    //
    // note that this will cause another buffer allocation
    // if there isn't really one available, but that is OK
    // since the caller needs the flag set again in order
    // to know to call us again
    //
    result = write_completed_handler();
    if( result && !interrupt_result )
        interrupt_result = result;
}



// called when main() wants to flush flows for a
// partially-filled block because it got queried
//
void AtmCardRunning::set_pending_block_usage( void )
{
    if( my_block )
    {
#if AMORTIZE_PCI_BUS_CYCLES && ATTEMPT_UNAMORTIZE
        // joel debug
        //
        static Bit32 current0;
        current0 = amcc->MasterWriteAddress;
        Bit32 current0a = current0;

        // should we use a local copy or the same class variable
        // that the interrupt handler uses???
        //
        AmccBusMasterControlStatus_register amcc_bus;

        // force the AMCC to flush the up-to-3 words
        // that it has in its write FIFO but which
        // aren't enough to make it request the PCI bus
        //
        amcc_bus = amcc->BusMasterControlStatus;
        amcc_bus.WriteFifo4FilledForPciReq = FALSE;
        amcc->BusMasterControlStatus = amcc_bus;
#endif

        Bit32 current = amcc->MasterWriteAddress;
        my_block->used = (current - flat_to_physical( my_block->buffers )) / BUFFER_SIZE;
#if PENDING_BLOCK_PARANOIA
        if( my_block->used > COMPL_RING_ENTRIES )
        {
            printf( "card 0x%.4X internal error: calculated %d usage in pending RX block; truncated to max %d\n", card_handle, my_block->used, COMPL_RING_ENTRIES );
            printf( "card 0x%.4X internal error: current=0x%.8X buffers=0x%.8X phys\n", card_handle, current, flat_to_physical(my_block->buffers) );
            my_block->used = COMPL_RING_ENTRIES;
        }
#endif

#if AMORTIZE_PCI_BUS_CYCLES && ATTEMPT_UNAMORTIZE
        // set the AMCC back to needing 4 words in its
        // write FIFO to cause it to request the PCI bus
        //
        amcc_bus.WriteFifo4FilledForPciReq = TRUE;
        amcc->BusMasterControlStatus = amcc_bus;
#endif
    }
}



// turn off warning about the unused assignment to rx_sts1
//
#pragma warn -aus

int AtmCardRunning::card_init( InitType itype, LoopbackSelection ltype, int want_rx, int want_tx )
{
    RxControl1_register rx_ctl1;
    RxControl2_register rx_ctl2;
    RxStatus1_register rx_sts1;
    RxMatchUnusedBits_register rx_match;
    int result;

    // check that caller wants us only to receive
    //
    if( !want_rx )
        return ERROR_POINT2_MUST_RX;
    if( want_tx )
        return ERROR_POINT2_CANNOT_TX;

    // assert S/UNI-622 reset (low power),
    // so it won't give us any data while we are discarding
    // data on the FPGA, so host doesn't get out of cell
    // alignment with data flowing from framer
    //
    suni->Master.Reset = TRUE;

    // assert external FIFO reset and load level
    //
    rx_ctl1.UnResetExternalFifo = FALSE;
    rx_ctl1.EnableCells = FALSE;
    rx_ctl1.EnableFifoWrite = FALSE;
    rx_ctl1.EnableAmccWrite = FALSE;
    rx_ctl1.UnLoadFifoLevels = FALSE;
    rx_ctl1.CopyTimestampsToHost = FALSE;
    rx_ctl1.CopyAtmHeadersToHost = FALSE;
    rx_ctl1.CopyAtmPayloadsToHost = FALSE;
    both_fpgas->Receive.Control1 = rx_ctl1;

    // assert AMCC FIFO reset
    //
    * (Bit32 *) &amcc_bus = 0;
    amcc_bus.ResetReadFifoFlags = TRUE;
    amcc_bus.ResetWriteFifoFlags = TRUE;
    amcc->BusMasterControlStatus = amcc_bus;

    // stop resetting the addon-to-PCI FIFO status flags,
    // enable card to write to host memory, give writes
    // (toward host) priority over reads (toward card),
#if AMORTIZE_PCI_BUS_CYCLES
    // and require 4 words to be in AMCC's write FIFO
    // before it requests the PCI bus, so as to have
    // longer bursts at the risk of not copying the last
    // up-to-12 bytes of the cell to the host
#else
    // and allow AMCC's to request the PCI bus when only
    // 1 word is in its write FIFO, so as to guarantee
    // that all words of a cell get copied to the host
#endif
    //
    amcc_bus.ResetReadFifoFlags = FALSE;
    amcc_bus.ResetWriteFifoFlags = FALSE;
    amcc_bus.EnableWriteTransfer = TRUE;
    amcc_bus.WriteVersusReadPriority = TRUE;
    amcc_bus.WriteFifo4FilledForPciReq = AMORTIZE_PCI_BUS_CYCLES;
//    amcc->BusMasterControlStatus = amcc_bus;

    // allocate a block of memory and give it to the card
    //
    // note that this routine also enables writes in the AMCC
    // bus master control/status register, so we delayed calling
    // it until we had initialized all other fields there
    //
    my_block = NULL;
    result = write_completed_handler();
    if( result )
    {
        printf( "card %d error %d allocating first buffer block\n", card_index, result );
        return result;
    }

    // enable AMCC to interrupt host when write transfer is complete
    //
    amcc->InterruptControlStatus.EnableWriteTransferCompleteInterrupt = TRUE;

    // stop resetting the external FIFO
    //
    rx_ctl1.UnResetExternalFifo = TRUE;
    both_fpgas->Receive.Control1 = rx_ctl1;

    // set the external FIFO's almost-empty level to 1
    // and its almost-full level to FULL-7
    //
    both_fpgas->Receive.FifoLevelWrite.Value = 1;
    both_fpgas->Receive.FifoLevelWrite.Value = 0;
    both_fpgas->Receive.FifoLevelWrite.Value = 7;
    both_fpgas->Receive.FifoLevelWrite.Value = 0;

    // stop loading FIFO levels
    //
    rx_ctl1.UnLoadFifoLevels = TRUE;
    both_fpgas->Receive.Control1 = rx_ctl1;

    // clear the latched RX errors by reading them
    //
    rx_sts1 = both_fpgas->Receive.Status1;

    // enable the parts of the cell we want copied to host
    //
    rx_ctl1.CopyTimestampsToHost = TRUE;
    rx_ctl1.CopyAtmHeadersToHost = TRUE;
    if( itype == PassIdleCells )
        rx_ctl1.CopyAtmPayloadsToHost = TRUE;
    else
        rx_ctl1.CopyAtmPayloadsToHost = CARD_COPIES_ATM_PAYLOAD;
    both_fpgas->Receive.Control1 = rx_ctl1;

    // unused bits of VPI/VCI (in this case the GFC)
    // must match 0
    //
    // DANGER: executing this write (at address +410h)
    // immediately after the write to RX control2 register
    // (at address +40Ch) allows the 82439HX to merge the
    // 2 accesses into a PCI burst which a POINT CPLD
    // bug handles very badly, causing target abort, host
    // retry, target abort ad infinitum --- hard lockup
    //
    rx_match.Value = 0;
    rx_match.Mask = 15;
    both_fpgas->Receive.MatchUnusedBits = rx_match;

    // UNI addressing
    //
    rx_ctl2.NumVciBitsStolenForVpi = 0;

    // pass all cells to host if desired
    //
    if( itype == PassIdleCells )
    {
        rx_ctl2.PassAllNonLastCells = TRUE;
        rx_ctl2.PassAllLastCells = TRUE;
        rx_ctl2.PassOamRmCells = TRUE;
    }
    else
    {
        rx_ctl2.PassAllNonLastCells = CARD_COPIES_ALL_CELLS | !CARD_COPIES_ATM_PAYLOAD;
        rx_ctl2.PassAllLastCells = CARD_COPIES_LAST_CELL | CARD_COPIES_ALL_CELLS | !CARD_COPIES_ATM_PAYLOAD;
        rx_ctl2.PassOamRmCells = CARD_COPIES_OAM_CELLS | CARD_COPIES_ALL_CELLS | !CARD_COPIES_ATM_PAYLOAD;
    }
    rx_ctl2.NumNonLastCellsToPass = CELLS_PER_PACKET;
    both_fpgas->Receive.Control2 = rx_ctl2;

    // clear the RX FPGA's counts of various kinds of cells,
    // then its latch, then do it one more time because the
    // compiler wants to optimize out the last assignment
    // to a variable which is thereafter unused
    //
    both_fpgas->Receive.CellCountTransferReset.LatchAndClear = TRUE;
    both_fpgas->Receive.CellCountTransferReset.LatchAndClear = TRUE;
    both_fpgas->Receive.CellCountTransferReset.LatchAndClear = TRUE;

    // tell the TriQuint serial/parallel chip how to loop
    //
    aim_and_cpld->Aim.LoopbackControl.Select = ltype;

    // enable utopia cell reception
    //
    rx_ctl1.EnableCells = TRUE;
    both_fpgas->Receive.Control1 = rx_ctl1;

    // enable writes to AMCC FIFO
    //
    rx_ctl1.EnableAmccWrite = TRUE;
    both_fpgas->Receive.Control1 = rx_ctl1;

    // enable writes to external FIFO
    //
    rx_ctl1.EnableFifoWrite = TRUE;
    both_fpgas->Receive.Control1 = rx_ctl1;

    // deassert S/UNI-622 reset
    //
    // note that if the framer was already enabled, this would
    // clear all S/UNI counters, including RX section BIP-8 error count
    //
    suni->Master.Reset = FALSE;

    // remember that we need cleanup during destructor
    //
    card_was_initialized = TRUE;

    // let our (someday) subclass determine how to initialize the S/UNI 622
    //
    result = card_init_suni( itype );
    if( result )
        return result;

    return 0;
}

// turn on warnings about unused variable assignments
// if they were on initially
//
// this must be done after the end of the function,
// not immediately after the assignment itself
//
#pragma warn .aus



int AtmCardRunning::card_init_suni( InitType mode )
{
    SuniRacpControl_register suni_ctl;
    SuniGfcAndMiscControl_register suni_ppp;

    suni_ctl.DisablePayloadDescrambling = FALSE;
    suni_ctl.ResetFifo                  = FALSE;
    suni_ctl.AddCosetPolynomialToHcs    = TRUE;
    suni_ctl.PassUncorrectableHcsErrors = FALSE;
    suni_ctl.DisableHcsErrorCorrection  = FALSE;
    suni_ctl.RxParityEven               = FALSE;
    suni_ctl.EnableFixedStuffing        = TRUE;
    if( mode == PassIdleCells )
        suni_ctl.PassFilteredCells      = TRUE;
    else
        suni_ctl.PassFilteredCells      = FALSE;
    suni->RacpControl = suni_ctl;

    suni_ppp.RxGfcEnable0               = TRUE;
    suni_ppp.RxGfcEnable1               = TRUE;
    suni_ppp.RxGfcEnable2               = TRUE;
    suni_ppp.RxGfcEnable3               = TRUE;
    suni_ppp.Unused                     = 0;
    suni_ppp.RxByteParity               = FALSE;
    if( mode == NormalPpp )
        // force cell delineation always (PPP mode)
        //
        suni_ppp.CellDelineationDisable = TRUE;
    else    // assume mode == NormalAtm
        // allow SUNI to search for cell delineation (ATM mode)
        //
        suni_ppp.CellDelineationDisable = FALSE;
    suni->GfcAndMiscControl = suni_ppp;

    return 0;
}



// turn off warning that the loop waiting for RX FPGA to
// latch and clear counters "has no effect"
//
#pragma warn -eff

void AtmCardRunning::card_exit( void )
{
    RxControl1_register rx_ctl1;
    RxStatus1_register rx_sts1;
    Bit32 temp;
    int result;

    // tell the RX FPGA not to read SUNI or write to external FIFO
    // or write to AMCC FIFO
    //
    rx_ctl1.UnResetExternalFifo = FALSE;
    rx_ctl1.EnableCells = FALSE;
    rx_ctl1.EnableFifoWrite = FALSE;
    rx_ctl1.EnableAmccWrite = FALSE;
    rx_ctl1.UnLoadFifoLevels = FALSE;
    rx_ctl1.CopyTimestampsToHost = FALSE;
    rx_ctl1.CopyAtmHeadersToHost = FALSE;
    rx_ctl1.CopyAtmPayloadsToHost = FALSE;
    both_fpgas->Receive.Control1 = rx_ctl1;

    // display the AMCC's idea of who interrupted
    //
    if( amcc_intr.Interrupted )
        fprintf( stdout, "AMCC thinks it interrupted at all\n" );
    if( amcc_intr.ReadTransferCompleted )
        fprintf( stdout, "AMCC thinks read transfer completed\n" );
    if( amcc_intr.WriteTransferCompleted )
        fprintf( stdout, "AMCC thinks write transfer completed\n" );
    if( amcc_intr.IncomingMailBoxInterrupted )
        fprintf( stdout, "AMCC thinks incoming mailbox interrupted host\n" );
    if( amcc_intr.OutgoingMailBoxInterrupted )
        fprintf( stdout, "AMCC thinks outgoing mailbox interrupted host\n" );
    if( amcc_intr.TargetAborted )
        fprintf( stdout, "AMCC thinks PCI bus target aborted\n" );
    if( amcc_intr.MasterAborted )
        fprintf( stdout, "AMCC thinks PCI bus master aborted\n" );

    // disable AMCC from interrupting host
    //
    amcc_intr.EnableWriteTransferCompleteInterrupt = FALSE;
    amcc_intr.EnableReadTransferCompleteInterrupt = FALSE;
    amcc_intr.EnableOutgoingMailBoxGoesEmptyInterrupt = FALSE;
    amcc_intr.EnableIncomingMailBoxGoesFullInterrupt = FALSE;
    amcc->InterruptControlStatus = amcc_intr;

    // disable AMCC writing to host memory
    //
    amcc->BusMasterControlStatus.EnableWriteTransfer = FALSE;

    // display how many interrupts have occurred
    //
    fprintf( stdout, "card %d interrupted %u times\n", card_index, interrupt_count );

    // display AMCC write address and remaining count
    //
    fprintf( stdout, "AMCC next write to host would be at 0x%.8X/0x%.8X log/phys\n",
        physical_to_flat( amcc->MasterWriteAddress ), amcc->MasterWriteAddress );
    fprintf( stdout, "AMCC could write %d more bytes to host out of original %d\n",
        amcc->MasterWriteTransferCount, COMPL_RING_ENTRIES * BUFFER_SIZE );

#if CARD_EXIT_FILLS_PARTIAL_BLOCK
    // count any cells left over in pending block, which were not
    // quite enough to cause a pending block full interrupt
    //
    // must do this before zeroing MasterWriteAddress pointer on card!
    //
    set_pending_block_usage();
#endif

    // tell the AMCC there isn't any memory left on host to write to
    //
    amcc->MasterWriteTransferCount = 0;

    // latch the RX FPGA's counts of various kinds of cells
    // then wait for the card to finish that job
    //
    // this needs to use a watchdog timer!
    //
    both_fpgas->Receive.CellCountTransferReset.LatchAndClear = TRUE;
    while( both_fpgas->Receive.CellCountTransferReset.LatchAndClear )
        ;

    // display the RX FPGA counts of various cell types
    //
    temp = (both_fpgas->Receive.OamRmCellCountRead0.Value)
         | (both_fpgas->Receive.OamRmCellCountRead1.Value << 8)
         | (both_fpgas->Receive.OamRmCellCountRead2.Value << 16)
         | (both_fpgas->Receive.OamRmCellCountRead3.Value << 24);
    fprintf( stdout, "%d OAM/RM cells seen by RX FPGA\n", temp );

    temp = (both_fpgas->Receive.NniMismatchCountRead0.Value)
         | (both_fpgas->Receive.NniMismatchCountRead1.Value << 8)
         | (both_fpgas->Receive.NniMismatchCountRead2.Value << 16)
         | (both_fpgas->Receive.NniMismatchCountRead3.Value << 24);
    fprintf( stdout, "%d cells didn't match RX FPGA's NNI expectations\n", temp );

    // display the external FIFO statuses and AIM parity error
    //
    rx_sts1 = both_fpgas->Receive.Status1;
    if( rx_sts1.FifoEmpty )
        fprintf( stdout, "RX FPGA thinks external FIFO is empty now\n" );
    if( rx_sts1.FifoAlmostEmpty )
        fprintf( stdout, "RX FPGA thinks external FIFO is almost empty now\n" );
    if( rx_sts1.FifoAlmostFull )
        fprintf( stdout, "RX FPGA thinks external FIFO is almost full now\n" );
    if( rx_sts1.FifoFull )
        fprintf( stdout, "RX FPGA thinks external FIFO is full now\n" );
    if( rx_sts1.FifoEverAlmostFull )
        fprintf( stdout, "RX FPGA thinks external FIFO was ever almost full\n" );
    if( rx_sts1.FifoEverFull )
        fprintf( stdout, "RX FPGA thinks external FIFO was ever full\n" );
    if( rx_sts1.AimEverRxParityError )
        fprintf( stdout, "RX FPGA thinks AIM RX parity error ever occurred\n" );

    // latch and display SUNI622 counts
    // (not implemented yet)
    //

#if CARD_EXIT_FILLS_PARTIAL_BLOCK
    // if there is a non-empty block left over...
    //
    if( my_block && my_block->used>0 )
    {
        // place my_block on global filled list
        // even though it is not quite all filled
        //
        result = all_buffers->fill( my_block );
        if( !result )
            my_block = NULL;
    }
#endif

#if CARD_EXIT_RELEASES_BLOCKS
    if( my_block )
    {
        my_block->state = is_consuming;     // force free to succeed
        all_buffers->free( my_block );
        my_block = NULL;
    }
#endif
}

// turn on warnings about code having no effect
// if they were on initially
//
// this must be done after the end of the function,
// not immediately after the assignment itself
//
#pragma warn .eff



