#include <stdio.h>                      // printf()
#include "../intel/timestmp.h"          // PentiumClock
#include "../apptel/amcc.h"             // Amcc_registers
#include "../apptel/nvram.h"            // prototypes we must adhere to



// manifest constants
//
#define MAX_CYCLE_DELAY     1.0         // in seconds



// global variables for statistics
//
static PentiumClock timing_min, timing_max, timing_tot;
static Bit32 iterations_min, iterations_max, iterations_tot;
static Bit32 num_calls;



// initialize statistics
//
void nvram_init_stats( void )
{
    num_calls = 0;

    timing_min.low = timing_min.high = BIT32_MAX;
    timing_max = (Bit32) 0;
    timing_tot = (Bit32) 0;

    iterations_min = BIT32_MAX;
    iterations_max = 0;
    iterations_tot = 0;
}



void nvram_print_stats( void )
{
    if( !num_calls )
    {
        fprintf( stdout, "nvram_ready_wait() never called successfully\n" );
        return;
    }

    fprintf( stdout, "nvram_ready_wait() called %u times successfully\n", num_calls );

    fprintf( stdout, "timing: min/max/mean %g/%g/%g uS\n",
        1e6 * timing_min.get_seconds(),
        1e6 * timing_max.get_seconds(),
        1e6 * timing_tot.get_seconds() / num_calls
    );

    fprintf( stdout, "iterations: min/max/mean %u/%u/%g\n",
        iterations_min, iterations_max,
        (double)iterations_tot / num_calls
    );
}



// wait for the AMCC nvRAM interface to be ready (not busy )
//
static int nvram_ready_wait(
    Amcc_registers *amcc,
    AmccBusMasterControlStatus_register *control_status
)
{
    PentiumClock now, alarm, before;
    Bit32 iterations;
    int cycle_is_pending;

    alarm.put_seconds( MAX_CYCLE_DELAY );
    before.set();
    alarm += before;
    iterations = 0;
    do
    {
        *control_status = amcc->BusMasterControlStatus;
        now.set();
        iterations++;

    } while( control_status->NvramBusy && now<=alarm );

    if( control_status->NvramBusy )
    {
        fprintf( stderr, "error: more than %f seconds (%u iterations) elapsed"
            " waiting for NVRAM cycle to complete\n", iterations, MAX_CYCLE_DELAY );
        return ERROR_NVRAM_CYCLE_TIMEOUT;
    }

    // update our stats
    //
    num_calls++;

    iterations_tot += iterations;
    if( iterations_max < iterations )
        iterations_max = iterations;
    if( iterations_min > iterations )
        iterations_min = iterations;

    now -= before;
    timing_tot += now;
    if( timing_max < now )
        timing_max = now;
    if( timing_min > now )
        timing_min = now;

    return 0;
}



extern "C"
int nvram_read_or_write( Amcc_registers *amcc, Bit16 address, int want_write, Bit8 *data0 )
{
    AmccBusMasterControlStatus_register control_status;
    int result;

    if( !amcc )
        return ERROR_NVRAM_NOT_INITIALIZED;

    if( address > MAX_NVRAM_ADDRESS )
    {
        fprintf( stderr, "address 0x%.8X is larger than maximum 0x%.8X\n", address, MAX_NVRAM_ADDRESS );
        return ERROR_NVRAM_INVALID_ADDRESS;
    }

    // initial wait for AMCC nvRAM i/f readiness
    //
    result = nvram_ready_wait( amcc, &control_status );
    if( result )
        return result;

    // load low byte of address
    //
    control_status.NvramBusy = TRUE;
    control_status.NvramAccessType = LoadLowAddressByte;
    control_status.NvramAddressOrData = address & 255;
    amcc->BusMasterControlStatus = control_status;
    result = nvram_ready_wait( amcc, &control_status );
    if( result )
        return result;

    // load high byte of address
    //
    control_status.NvramBusy = TRUE;
    control_status.NvramAccessType = LoadHighAddressByte;
    control_status.NvramAddressOrData = address >> 8;
    amcc->BusMasterControlStatus = control_status;
    result = nvram_ready_wait( amcc, &control_status );
    if( result )
        return result;

    if( want_write )
    {
        // load data byte and begin write
        //
        control_status.NvramBusy = TRUE;
        control_status.NvramAccessType = BeginWrite;
        control_status.NvramAddressOrData = *data0;
        amcc->BusMasterControlStatus = control_status;
        result = nvram_ready_wait( amcc, &control_status );
        if( result )
            return result;
    }
    else
    {
        // begin read
        //
        control_status.NvramBusy = TRUE;
        control_status.NvramAccessType = BeginRead;
        control_status.NvramAddressOrData = 0xA5;       // helps detect failure
        amcc->BusMasterControlStatus = control_status;
        result = nvram_ready_wait( amcc, &control_status );
        if( result )
            return result;

        // copy the data to the caller's variable
        //
        *data0 = control_status.NvramAddressOrData;
    }

    return 0;
}



