#include <stdio.h>                      // printf()
#include <stdlib.h>                     // atoi() strtoul()
#include "../apptel/point.h"            // class AtmCard
#include "../apptel/sramtest.h"         // prototypes we must adhere to



// conditional compile switches
//
#define SRAM_WALK_VERBOSE       TRUE



// global variables
//
AtmCard *card;



static int sram_read_after_write( Bit32 address, Bit32 new_value, const char *state, int error_base )
{
    Bit32 value_read;
    int result;

#if SRAM_WALK_VERBOSE
    fprintf( stdout, " 0x%.8X", new_value );
#endif

    result = sram_read_or_write( &card->both_fpgas->Receive, address, TRUE, &new_value );
    if( result )
    {
        fprintf( stderr, "error %d writing %s 0x%.8X to SRAM address 0x%.8X\n", result, state, new_value, address );
        return error_base;
    }

    value_read = 123;

    result = sram_read_or_write( &card->both_fpgas->Receive, address, FALSE, &value_read );
    if( result )
    {
        fprintf( stderr, "error %d reading %s 0x%.8X from SRAM address 0x%.8X\n", result, state, new_value, address );
        return error_base + 1;
    }

    if( value_read != new_value )
    {
        fprintf( stderr, "read 0x%.8X instead of %s 0x%.8X at SRAM address 0x%.8X\n", value_read, state, new_value, address );
        return error_base + 2;
    }

    return 0;
}



static int sram_write_after_read( Bit32 address, Bit32 old_value, Bit32 new_value, const char *state, int error_base )
{
    Bit32 value_read;
    int result;

    value_read = 123;
    result = sram_read_or_write( &card->both_fpgas->Receive, address, FALSE, &value_read );
    if( result )
    {
        fprintf( stderr, "error %d reading %s 0x%.8X from SRAM address 0x%.8X\n", result, state, old_value, address );
        return error_base + 1;
    }

    if( value_read != old_value )
    {
        fprintf( stderr, "read 0x%.8X instead of %s 0x%.8X at SRAM address 0x%.8X\n", value_read, state, old_value, address );
        return error_base + 2;
    }

    result = sram_read_or_write( &card->both_fpgas->Receive, address, TRUE, &new_value );
    if( result )
    {
        fprintf( stderr, "error %d writing %s 0x%.8X to SRAM address 0x%.8X\n", result, state, new_value, address );
        return error_base;
    }

    return 0;
}



static int sram_walk_1_thru_0( void )
{
    Bit32 count;
    int result;
    Bit32 sram_data;

    printf( "walking 1 through 0's at address 0\n" );

    // initial 0's
    //
    result = sram_read_after_write( 0, 0, "initial", ERROR_SRAM_TEST_INIT_0_WRITE );
    if( result )
        return result;

    // walk 1 through 0's
    //
    sram_data = 1;
    for( count=0; count<32; count++ )
    {
        result = sram_read_after_write( 0, sram_data, "walking", ERROR_SRAM_TEST_WALK_1_WRITE );
        if( result )
            return result;

        sram_data <<= 1;
    }

    // final 0's
    //
    result = sram_read_after_write( 0, 0, "final", ERROR_SRAM_TEST_FINAL_0_WRITE );
#if SRAM_WALK_VERBOSE
    fprintf( stdout, "\n" );
#endif
    if( result )
        return result;

    return 0;
}



static int sram_walk_0_thru_1( void )
{
    Bit32 count;
    int result;
    Bit32 sram_data;

    printf( "walking 0 through 1's at address 0\n" );

    // initial 1's
    //
    result = sram_read_after_write( 0, -1, "initial", ERROR_SRAM_TEST_INIT_1_WRITE );
    if( result )
        return result;

    // walk 0 through 1's
    //
    sram_data = ~1;
    for( count=0; count<32; count++ )
    {
        result = sram_read_after_write( 0, sram_data, "walking", ERROR_SRAM_TEST_WALK_0_WRITE );
        if( result )
            return result;

        sram_data = (sram_data << 1) | 1;
    }

    // final 1's
    //
    result = sram_read_after_write( 0, -1, "final", ERROR_SRAM_TEST_FINAL_1_WRITE );
#if SRAM_WALK_VERBOSE
    fprintf( stdout, "\n" );
#endif
    if( result )
        return result;

    return 0;
}



#define SRAM_CHECKERBOARD_PATTERN_1         0x55555555
#define SRAM_CHECKERBOARD_PATTERN_2         0xAAAAAAAA

static int sram_checkerboard( void )
{
    Bit32 sram_address;
    Bit32 count;
    Bit32 sram_data;
    int result;

    // write initial PATTERN_1 to all locations, using ascending addresses
    //
    fprintf( stdout, "checkerboard pass 1: writing 0x%.8X to ascending addresses\n",
        SRAM_CHECKERBOARD_PATTERN_1 );
    sram_data = SRAM_CHECKERBOARD_PATTERN_1;
    for( sram_address=0; sram_address<=MAX_SRAM_ADDRESS; sram_address++ )
    {
        result = sram_read_or_write( &card->both_fpgas->Receive, sram_address, TRUE, &sram_data );
        if( result )
        {
            fprintf( stderr, "error %d writing checkerboard pass 1 pattern 1"
                " = 0x%.8X to SRAM address 0x%.8X\n", result, sram_data, sram_address );
            return ERROR_SRAM_TEST_CHK_WRITE_1;
        }
    }

    // check each location for PATTERN_1 while changing to PATTERN_2,
    // using ascending addresses
    //
    fprintf( stdout, "checkerboard pass 2: reading ascending addresses and comparing"
        " to 0x%.8X while writing 0x%.8X\n", SRAM_CHECKERBOARD_PATTERN_1,
        SRAM_CHECKERBOARD_PATTERN_2 );
    for( sram_address=0; sram_address<=MAX_SRAM_ADDRESS; sram_address++ )
    {
        result = sram_write_after_read( sram_address,
            SRAM_CHECKERBOARD_PATTERN_1, SRAM_CHECKERBOARD_PATTERN_2,
            "checkerboard pass 2", ERROR_SRAM_TEST_CHK_WRITE_2 );
        if( result )
            return result;
    }

    // check each location for PATTERN_2 while changing back to
    // PATTERN_1, using descending addresses
    //
    // assumes the MAX_SRAM_ADDRESS < BIT32_MAX, because the loop
    // termination condition assumes it can detect underflow as
    // a wrap to a value greater than MAX_SRAM_ADDRESS
    //
    fprintf( stdout, "checkerboard pass 3: reading descending addresses and comparing"
        " to 0x%.8X while writing 0x%.8X\n", SRAM_CHECKERBOARD_PATTERN_2,
        SRAM_CHECKERBOARD_PATTERN_1 );
    for( sram_address=MAX_SRAM_ADDRESS; sram_address<=MAX_SRAM_ADDRESS; sram_address-- )
    {
        result = sram_write_after_read( sram_address,
            SRAM_CHECKERBOARD_PATTERN_2, SRAM_CHECKERBOARD_PATTERN_1,
            "checkerboard pass 3", ERROR_SRAM_TEST_CHK_WRITE_3 );
        if( result )
            return result;
    }

    return 0;
}



int sram_test( void )
{
    int result;

    result = sram_walk_1_thru_0();
    if( result )
        return result;

    result = sram_walk_0_thru_1();
    if( result )
        return result;

    return sram_checkerboard();
}



int main( int argc, char *argv[] )
{
    Bit16 card_index;
    int result;
    const char *program_name = argv[0];

    // parse command-line arguments
    //
    if( argc != 2 )
    {
        fprintf( stderr, "tests all SRAM words on the Applied Telecom PCI card\n\n" );
        fprintf( stderr, "usage: %s cardindex\n", program_name );
        return ERROR_SRAM_MISSING_ARGS;
    }

    card_index = atoi( argv[1] );

    card = new AtmCard( card_index );
    if( !card )
        return ERROR_SRAM_CANNOT_ALLOC;
    if( card->init_result )
        return card->init_result;

    sram_init_stats();

    result = sram_test();

    sram_print_stats();

    delete card;

    if( result )
        return result;

    return 0;
}
