// this source file should be linked with CAPTURE.CPP
// instead of FS2FLOWS.C and DOSTATS.C and
// PORTSTAT.C and NETSTATS.C and HOSTSTAT.C and
// UTIL.C and READBGP.C and SUBNET.C and QUEUE.C
//



#include <string.h>                         // strcpy()
#include "../intel/byteswap.h"              // ntohl()
#include "../integrat/buffer.h"             // class BufferBlock class BuffferQueue
#include "../integrat/fs2flows.h"           // prototypes we must adhere to



// manifest constants
//
#define MAX_FILENAME_SIZE               1000
#define TRACE_ALREADY_SWAPPED           TRUE



// global variables
//
int want_disk_io_during_run = FALSE;
int file_is_open = FALSE;
int saw_eof = FALSE;
char filename[ MAX_FILENAME_SIZE ];
FILE *in_handle;
Bit32 block_count = 0;
Signed16 cardindex_wanted;
Bit32 wrong_cardindex_cells = 0;
Bit32 valid_cells = 0;
BufferBlock *block = 0;



// note: flows_per_block() is not implemented
// because AAL5.C function cells_reassembly() won't be called
//



// turn off compiler warning about unused parameters
// since I know they won't be used and I must have
// them because calling code cannot be changed
//
#pragma warn -par

int flows_per_block( BufferBlock *block )
{
    return 0;
}

// turn compiler warning about unused parameters
// back to whatever it was before I messed with it
//
#pragma warn .par



// turn off compiler warning about unused parameters
// since I know they won't be used and I must have
// them because calling code cannot be changed
//
#pragma warn -par

int flows_init( int argc, char *argv[], PentiumClock *keystroke_interval,
    PentiumClock *query_interval, PentiumClock *active_interval,
    int *want_facility_loopback, int *want_rx, int *want_tx )
{
    int result;
    const char *program = argv[ 0 ];
    const char *filename0 = argv[ 1 ];
    const char *cardstr = argv[ 2 ];

    if( argc!=3 && argc!=4 )
    {
        fprintf( stderr, "\n" );
        fprintf( stderr, "usage: %s filename cardindex {placeholder}\n", program );
        fprintf( stderr, "where -1 cardindex means all cards\n" );
        fprintf( stderr, "and anything in third parm causes disk read during sending, instead of before\n" );
        fprintf( stderr, "\n" );
        return 1;
    }

    result = sscanf( cardstr, "%d", &cardindex_wanted );
    if( result != 1 )
    {
        fprintf( stderr, "unable to convert cardindex '%s' to integer\n", cardstr );
        return 4;
    }

    // we cannot call any BufferQueue member functions
    // because the global pointer "all_buffers" has not
    // been assigned to yet
    //
    // so remember the filename and open the file later
    // the first time we need it
    //
    strcpy( filename, filename0 );
    file_is_open = FALSE;

    if( argc == 4 )
        want_disk_io_during_run = TRUE;

    if( want_disk_io_during_run )
        fprintf( stderr, "\nsending preformatted cells from disk until EOF, then exiting\n\n" );
    else
        fprintf( stderr, "\nreading preformatted cells until host memory is full, then stopping and sending them all\n\n" );

    // capture module should enable TX but not RX on all cards
    //
    *want_rx = FALSE;
    *want_tx = TRUE;

    // cannot have loopback on if we want to transmit our own data
    // (I presume)
    //
    *want_facility_loopback = FALSE;

    // capture module should ask me about query every chance it gets
    //
    query_interval->put_seconds( 0 );

    return 0;
}

// turn compiler warning about unused parameters
// back to whatever it was before I messed with it
//
#pragma warn .par



int flows_cleanup( void )
{
    return 0;
}




// called by main() many times during run
//
int flows_is_query_wanted( void )
{
    if( !file_is_open || !saw_eof || want_disk_io_during_run )
        return TRUE;

    return FALSE;
}


int flows_is_query_pending( void )
{
    return FALSE;
}



// I know the extra_msg paramater isn't used,
// and I don't want the compiler to nag me about it
//
#pragma warn -par

// opens disk file the first time it is called
//
// assumes that main() will start transmission if anything is on
// the sendable list when we return, so...
//
// if the user doesn't want disk I/O during run,
// tries to read all blocks before returning,
// and will stop file read when block allocate fails
// and will do nothing on later calls
//
// if the user wants disk I/O during run,
// only reads one block each time it is called,
// and will retry a failed block allocate on later calls
//
int flows_query( const char *extra_msg )
{
    int result;
    Bit8 header[ SECTOR_SIZE ];
    traceblock_send *buffer = (traceblock_send *) header;
    tracerec_send *record;
    int file_was_open = file_is_open;
    unsigned count;
    Bit32 used;

    if( !file_is_open )
    {
        // try to open the file
        //
        in_handle = fopen( filename, "rb" );
        if( in_handle == NULL )
        {
            fprintf( stderr, "error %d opening input file\n", errno );
            return 2;
        }

        fprintf( stderr, "BufferBlock's will be read from file '%s' %s\n",
            filename,
            want_disk_io_during_run ? "later" : "now"
        );

        file_is_open = TRUE;
    }

    if( !saw_eof && (want_disk_io_during_run || !file_was_open) )
    while( 1 )
    {
        if( feof( in_handle ) || ferror( in_handle ) )
        {
            saw_eof = TRUE;
            fclose( in_handle );
            fprintf( stderr, "\n" );
            if( ferror( in_handle ) )
                return 3;
            break;
        }

        block_count++;

        if( block == NULL )
        {
            result = all_buffers->allocate( &block );
            if( result )
            {
                fprintf( stderr, "\n" );
                if( result == ERROR_QUEUE_NOT_FREE )
                {
                    // forget that this failed, so main() doesn't try
                    // to pass the next free buffer to the cards for receive
                    //
                    all_buffers->has_allocate_failed_lately();

                    // don't cause program exit
                    //
                    return 0;
                }

                fprintf( stderr, "unexpected error %d allocating block %d; aborting\n", result, block_count );
                return result;
            }
        }

        // read the next block header
        //
        result = fread( header, sizeof(header), 1, in_handle );
        if( result != 1 )
        {
            saw_eof = TRUE;
            block_count--;
            fprintf( stderr, "\n" );
            if( !feof( in_handle ) )
            {
                fprintf( stderr, "error %d reading input file block %d header\n", errno, block_count+1 );
                return 5;
            }
            fclose( in_handle );
            break;
        }

        // read the trace records into the BufferBlock
        //
        // note that this must NOT be conditional on our
        // liking the block, or we will be unable to read
        // the next block!
        //
        result = fread( block->buffers, SEND_RING_ENTRIES*sizeof(tracerec_send), 1, in_handle );
        if( result != 1 )
        {
            saw_eof = TRUE;
            fprintf( stderr, "\n" );
            if( !feof( in_handle ) )
                fprintf( stderr, "error %d reading input file block %d data\n", errno, block_count );

            fclose( in_handle );
            return 6;
        }

#if HOST_IS_BIG_ENDIAN != TRACE_ALREADY_SWAPPED
        // swap the usage count
        //
        buffer->used = ntohl( buffer->used );
#endif

        fprintf( stderr, " %d(%d)", block_count, buffer->cardindex );

        // must remember usage count
        //
        // at this point it is in host byte order
        //
        used = buffer->used;

        // check for out-of-bound record count
        //
        if( used > SEND_RING_ENTRIES )
        {
            fprintf( stderr, "\n" );
            fprintf( stderr, "error: usage count of block %u is %u, but max is %u!\n", block_count, used, SEND_RING_ENTRIES );
            break;
        }

        if( cardindex_wanted!=-1 && buffer->cardindex!=cardindex_wanted )
        {
            fprintf( stderr, "ignored" );
            wrong_cardindex_cells += used;
            if( want_disk_io_during_run )
                break;
            continue;
        }

        valid_cells += used;

        // copy the fields we like to the BufferBlock
        //
        block->cardnum = 0;
        block->used = buffer->used;
        block->queried = 0;
        block->freed = 0;

        // for all records in the BufferBlock
        //
        for( count=0; count<block->used; count++ )
        {
            record = (tracerec_send *) (block->buffers + count*SEND_BUFFER_SIZE);

#if HOST_IS_BIG_ENDIAN != TRACE_ALREADY_SWAPPED
            // swap the repeat count
            //
            record->count = ntohl( record->count );
#endif

            // the rest of the data is IP and TCP headers,
            // and was always in network byte order!
            //
        }

        // put this block on the sendable list for the desired card
        //
        result = all_buffers->fill_tx( block, buffer->cardindex );
        if( result )
        {
            fprintf( stderr, "\n" );
            fprintf( stderr, "error %d from BufferQueue::fill_tx()\n", result );
            return result;
        }

        // remember to allocate a new block on next iteration
        //
        block = NULL;

        if( want_disk_io_during_run )
        {
            fprintf( stderr, "\n" );
            break;
        }
    }

    if( !file_was_open )

        for( count=0; count<MAX_ATM_NICS; count++ )

            // trigger main() to keep trying to give the first buffer to this card
            //
            all_buffers->read_tx_failure_count[ count ]++;

    return 0;
}
#pragma warn .par



void flows_send_active( void )
{
}



// turn off compiler warning about unused parameters
// since I know they won't be used and I must have
// them because calling code cannot be changed
//
#pragma warn -par

// this code doesn't have anything special to do
// when keys are pressed
//
void flows_keystroke( int keystroke )
{
}
// turn compiler warning about unused parameters
// back to whatever it was before I messed with it
//
#pragma warn .par



void sock_init( void )
{
}



