#include <mem.h>                        // memset()
#include <stdio.h>                      // fopen()
#include <errno.h>                      // errno
#include <conio.h>                      // kbhit()
#include <string.h>                     // strcat()
#include "trace.h"
#include "../intel/byteswap.h"          // ntohs() ntohl()
#include "buffer.h"                     // BufferBlock
#include "fs2flows.h"                   // flows_init() flows_cleanup() flows_per_block()



// conditional compilation switches
//
#define DEBUG_BLOCK_LOCKUP              FALSE
#define ALLOW_KEYPRESS_TO_HALT          TRUE
#define TRACE_ALREADY_SWAPPED           TRUE
#define CELLS_DUMP_WHEN_ALLOC_FAILS     FALSE
#define FORCE_CELL_COUNT                FALSE
#if CARD_IS_OC12
#    define FIX_TIMESTAMP_HIGH          FALSE       // OC12 card is fine and has different fields anyways
#else
#    define FIX_TIMESTAMP_HIGH          TRUE
#endif



// global variables
//
PentiumClock fake_interval;
int fake_argc;
int fake_loopback;
char *fake_argv[ 200 ];
int fake_want_rx;
int fake_want_tx;

#if DEBUG_BLOCK_LOCKUP
    Bit32 block_count_trace = 24;
#endif



int main( int argc, char *argv[] )
{
    const char *program = argv[ 0 ];
    const char *infile  = argv[ 1 ];
    const char *cardstr = argv[ 2 ];
    FILE *in_handle;
    int result;
    Bit32 used;
    Bit32 count;
    Bit32 block_count = 0;
    Bit32 valid_cells = 0;
    Bit32 wrong_cardindex_cells = 0;
#if FIX_TIMESTAMP_HIGH
    Bit16 prev_when_low[ MAX_ATM_NICS ];
    Bit16 when_low;
    Bit32 when_high[ MAX_ATM_NICS ];
    Bit8 interface;
#else
    Bit32 when_temp;
#endif
    struct tracerec *packet;
    Signed16 cardindex_wanted;
    char id_string[ 1000 ];
    PentiumClock startup_time, before;
    PentiumClock elapsed, elapsed_min, elapsed_max, elapsed_tot;
    double elapsed_min2, elapsed_max2, elapsed_avg2;
    Bit8 header[ SECTOR_SIZE ];
    traceblock *buffer = (traceblock *) header;
    BufferBlock *block;

    fprintf( stdout, "each record is %d bytes\n", sizeof(struct tracerec) );
    fprintf( stdout, "ICMP alone accounts for %d bytes\n", sizeof(struct icmp) );
    fprintf( stdout, "%d records in a block\n\n", COMPL_RING_ENTRIES );

    if( argc < 3 )
    {
        fprintf( stderr, "calls flows entrypoints for blocks in an OCxMON trace file\n" );
        fprintf( stderr, "usage: %s infile cardindex {flows-args}\n", program );
        fprintf( stderr, "where -1 cardindex means all cards\n" );
        return 1;
    }

    in_handle = fopen( infile, "rb" );
    if( in_handle == NULL )
    {
        fprintf( stderr, "error %d opening input file\n", errno );
        return 2;
    }

    result = sscanf( cardstr, "%d", &cardindex_wanted );
    if( result != 1 )
    {
        fprintf( stderr, "unable to convert cardindex '%s' to integer\n", cardstr );
        return 4;
    }

    flat_init();

    CardClock::sync_time.set();

    // init TCP/IP library
    //
    // we must do this before any other physical memory
    // function calls because those must be undone in the
    // same order that they were done, and sock_init()
    // installs an atexit() handler which means we are not
    // in control of when sock_exit() gets called
    //
    sock_init();

    // get rid of our args
    //
    fake_argc = argc - 2;
    fake_argv[ 0 ] = argv[ 0 ];
    for( count=1; count<fake_argc; count++ )
        fake_argv[ count ] = argv[ count + 2 ];

    result = flows_init( fake_argc, fake_argv,
        &fake_interval, &fake_interval, &fake_interval,
        &fake_loopback, &fake_want_rx, &fake_want_tx );
    if( result )
    {
        printf( "error %d initializing flows\n", result );
        return 5;
    }

    // allocate all remaining flat memory to card buffer pool
    //
    // we must do this AFTER statistics initialize, so they
    // get first crack at flat memory, in case they are using
    // it instead of malloc()
    //
    all_buffers = new BufferQueue;
    if( all_buffers == NULL )
    {
        fprintf( stderr, "cannot allocate BufferQueue!\naborting\n" );
        return 6;
    }

#if FIX_TIMESTAMP_HIGH
    // clear overriding high 32 bits of timestamp
    // and knowledge of previous low 32 bits
    //
    for( interface=0; interface<MAX_ATM_NICS; interface++ )
    {
        when_high[ interface ] = 0;
        prev_when_low[ interface ] = 0;
    }
#endif


    // initialize our flows_per_block() timing statistics
    //
    elapsed_min.high = -1;
    elapsed_min.low  = -1;
    elapsed_max.high = 0;
    elapsed_max.low  = 0;
    elapsed_tot.high = 0;
    elapsed_tot.low  = 0;
    startup_time.set();

    while( !feof( in_handle ) && !ferror( in_handle )
#if ALLOW_KEYPRESS_TO_HALT
        && !kbhit()
#endif
    )
    {
        block_count++;

        result = all_buffers->allocate( &block );
        if( result )
        {
            fprintf( stderr, "error %d allocating block %d; aborting\n", result, block_count );
#if CELLS_DUMP_WHEN_ALLOC_FAILS
            cells_dump_all_packets( stdout );
#endif
            break;
        }

        // read the next block header
        //
        result = fread( header, sizeof(header), 1, in_handle );
        if( result != 1 )
        {
            if( !feof( in_handle ) )
                fprintf( stderr, "error %d reading input file block %d header\n", errno, block_count );

            block_count--;
            break;
        }

        // read the trace records into the BufferBlock
        //
        // note that this must NOT be conditional on our
        // liking the block, or we will be unable to read
        // the next block!
        //
        result = fread( block->buffers, COMPL_RING_ENTRIES*sizeof(tracerec), 1, in_handle );
        if( result != 1 )
        {
            if( !feof( in_handle ) )
                fprintf( stderr, "error %d reading input file block %d data\n", errno, block_count );

            break;
        }

#if HOST_IS_BIG_ENDIAN != TRACE_ALREADY_SWAPPED
        // swap the usage count and card number
        //
        buffer->used = ntohl( buffer->used );
        buffer->cardnum = ntohs( buffer->cardnum );
#endif

        fprintf( stderr, "%d(%d) ", block_count, buffer->cardindex );
#if DEBUG_BLOCK_LOCKUP
        if( block_count == block_count_trace )
            fprintf( stderr, "debugger wants to stop here\n" );
#endif

        // must remember usage count
        //
        // at this point it is in host byte order
        //
        used = buffer->used;

        // check for out-of-bound record count
        //
        if( used > COMPL_RING_ENTRIES )
        {
            fprintf( stderr, "error: usage count of block %u is %u, but max is %u!\n", block_count, used, COMPL_RING_ENTRIES );
            break;
        }

        if( cardindex_wanted!=-1 && buffer->cardindex!=cardindex_wanted )
        {
            fprintf( stderr, "ignoring cardnum 0x%.4X cardindex %d\n", buffer->cardnum, buffer->cardindex );
            wrong_cardindex_cells += used;
            continue;
        }

        valid_cells += used;

        // copy the fields we like to the BufferBlock
        //
#if FIX_TIMESTAMP_HIGH
        interface =
#endif
        block->cardindex = buffer->cardindex;
        block->cardnum = buffer->cardnum;
        block->used = buffer->used;
        block->queried = 0;

        // for all records in the BufferBlock
        //
        for( count=0; count<block->used; count++ )
        {
            packet = (tracerec *) (block->buffers + count*BUFFER_SIZE);

#if HOST_IS_BIG_ENDIAN != TRACE_ALREADY_SWAPPED
            // swap the bytes and words of the timestamp
            //
#if !FIX_TIMESTAMP_HIGH
            when_temp = ntohl( packet->when.low );
#endif
            packet->when.low  = ntohl( packet->when.high );
#if !FIX_TIMESTAMP_HIGH
            packet->when.high = when_temp;
#endif
#endif

#if FIX_TIMESTAMP_HIGH
            // deal with timer wrap
            //
            when_low = packet->when.low >> 16;
            if( when_low < prev_when_low[ interface ] )
                when_high[ interface ]++;
            prev_when_low[ interface ] = when_low;
            packet->when.high = when_high[ interface ];
#endif

#if FORCE_CELL_COUNT
            packet->when.low &= ~CELL_COUNT_MASK;
#endif

#if CARD_SWAPS_ATM_HEADER
#if HOST_IS_BIG_ENDIAN != TRACE_ALREADY_SWAPPED
            // swap the bytes of the ATM header
            //
            packet->atm.whole = ntohl( packet->atm.whole );
#endif
#endif

            // the rest of the data is IP and TCP headers,
            // and was always in network byte order!
            //
        }

        // mark the block as "being consumed for statistics"
        //
        block->state = is_consuming;

        // start the timer of flows processing
        //
        before.set();

        // hand it off to flows routines
        //
        result = flows_per_block( block );

        // stop the timer of flows processing
        //
        elapsed.set();

        // update the flows processing time statistics
        //
        elapsed -= before;
        elapsed_tot += elapsed;
        if( elapsed > elapsed_max )
            elapsed_max = elapsed;
        if( elapsed < elapsed_min )
            elapsed_min = elapsed;

        // check the result of flows processing
        //
        if( result )
        {
            fprintf( stderr, "error %d processing flows in block %d\n", result, block_count );
            break;
        }
    }
    fprintf( stderr, "\n" );

#if ALLOW_KEYPRESS_TO_HALT
    if( kbhit() )
    {
        getch();
        fprintf( stderr, "halted by keypress!\n" );
    }
#endif

    // finalize flows processing time statistics
    // and print them
    //
    elapsed.set();
    elapsed -= startup_time;
    if( block_count )
    {
        elapsed_min2 = elapsed_min.get_seconds() * 1000;
        elapsed_max2 = elapsed_max.get_seconds() * 1000;
        elapsed_avg2 = elapsed_tot.get_seconds() * 1000 / block_count;
        printf( "flows_per_block() took min %.4f max %.4f avg %.4f mS\n", elapsed_min2, elapsed_max2, elapsed_avg2 );
        printf( "%.2f seconds (%.2f%%) spent in flows_per_block() during run\n", elapsed_tot.get_seconds(), 100.0*(double)elapsed_tot/(double)elapsed );
    }
    else
        printf( "flows_per_block() was never called\n" );

    flows_cleanup();

    // wait for a query or a keystroke
    //
    fprintf( stderr, "waiting for query to come in or key to be pressed\n" );
    while( !flows_is_query_wanted() && !kbhit() )
        ;

    if( kbhit() )
        getch();
    else
    {
        id_string[ 0 ] = '\0';
        for( count=0; count<argc; count++ )
        {
            strcat( id_string, "'" );
            strcat( id_string, argv[count] );
            strcat( id_string, "' " );
        }
        strcat( id_string, "\n" );
        result = flows_query( id_string );
        if( result )
            printf( "error %d from flows query\n", result );

        // wait for all data to leave the machine
        //
        printf( "waiting for data to drain\n" );
        while( flows_is_query_pending() && !kbhit() )
            flows_is_query_wanted();
    }

    // show statistics
    //
    fprintf( stdout, "%lu blocks examined\n", block_count );
    if( cardindex_wanted != -1 )
        fprintf( stdout, "%lu cells without cardindex %d found\n", wrong_cardindex_cells, cardindex_wanted );
    fprintf( stdout, "%lu valid cells found\n", valid_cells );

    fclose( in_handle );

    flat_exit();

    // return successfully
    //
    return 0;
}

